# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""UI widgets for the Scene Settings tab."""

from __future__ import annotations

import logging
import os

from deadline.client.ui import block_signals
from . import blender_utils
from .template_filling import BlenderSubmitterUISettings
from . import ocio_utils as ocio
from qtpy.QtCore import QRegularExpression, QSize, Qt  # type: ignore
from qtpy.QtGui import QRegularExpressionValidator  # type: ignore
from qtpy.QtWidgets import (  # type: ignore
    QCheckBox,
    QComboBox,
    QFileDialog,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QWidget,
)

_logger = logging.getLogger(__name__)

# Default values for combo boxes in the addon UI.
COMBO_DEFAULT_ALL_RENDERABLE_CAMERAS = "All Renderable Cameras"
COMBO_DEFAULT_ALL_RENDERABLE_LAYERS = "All Renderable Layers"


class SceneSettingsWidget(QWidget):
    """Widget containing all top level scene settings."""

    def __init__(self, initial_settings: BlenderSubmitterUISettings, parent=None):
        super().__init__(parent=parent)

        self.dev_options = os.environ.get("DEADLINE_ENABLE_DEVELOPER_OPTIONS", "").upper() == "TRUE"

        # Save the two lists of selectable cameras.
        self.all_layer_selectable_cameras = initial_settings.all_layer_selectable_cameras
        self.current_layer_selectable_cameras = initial_settings.current_layer_selectable_cameras

        self._build_ui()
        self._load(initial_settings)

    def _build_ui(self):
        """Set up the UI."""
        layout = QGridLayout(self)

        qt_pos_index = 0
        self.proj_path_txt = QLineEdit(self)
        layout.addWidget(QLabel("Project Path"), qt_pos_index, 0)
        layout.addWidget(self.proj_path_txt, qt_pos_index, 1)
        self.proj_path_txt.setReadOnly(True)

        qt_pos_index += 1
        self.op_dir_txt = FileSearchLineEdit(directory_only=True)
        layout.addWidget(QLabel("Output Directory"), qt_pos_index, 0)
        layout.addWidget(self.op_dir_txt, qt_pos_index, 1)

        qt_pos_index += 1
        self.op_file_txt = QLineEdit(self)
        layout.addWidget(QLabel("Output File Prefix"), qt_pos_index, 0)
        layout.addWidget(self.op_file_txt, qt_pos_index, 1)

        qt_pos_index += 1
        self.scenes_box = QComboBox(self)
        layout.addWidget(QLabel("Scene"), qt_pos_index, 0)
        layout.addWidget(self.scenes_box, qt_pos_index, 1)

        qt_pos_index += 1
        self.render_engine_box = QComboBox(self)
        engine_items = [
            ("eevee", "eevee"),
            ("cycles", "cycles"),
            ("workbench", "workbench"),
        ]
        for engine_value, text in engine_items:
            self.render_engine_box.addItem(text, engine_value)
        layout.addWidget(QLabel("Render Engine"), qt_pos_index, 0)
        layout.addWidget(self.render_engine_box, qt_pos_index, 1)

        qt_pos_index += 1
        self.layers_box = QComboBox(self)
        self.scenes_box.currentIndexChanged.connect(self._fill_view_layers_box)
        layout.addWidget(QLabel("View Layers"), qt_pos_index, 0)
        layout.addWidget(self.layers_box, qt_pos_index, 1)

        qt_pos_index += 1
        self.scenes_box.currentIndexChanged.connect(self._fill_cameras_box)
        self.cameras_box = QComboBox(self)
        layout.addWidget(QLabel("Cameras"), qt_pos_index, 0)
        layout.addWidget(self.cameras_box, qt_pos_index, 1)

        qt_pos_index += 1
        self.enable_gpu_check = QCheckBox("Cycles GPU Rendering", self)
        self.gpu_device_box = QComboBox(self)
        # Natively supported GPU devices for Blender 3.6 and 4.2:
        # https://docs.blender.org/manual/en/3.6/render/cycles/gpu_rendering.html
        # https://docs.blender.org/manual/en/4.2/render/cycles/gpu_rendering.html
        gpu_device_options = ["CUDA", "OptiX", "HIP", "oneAPI", "Metal"]
        for device in gpu_device_options:
            self.gpu_device_box.addItem(device, device.upper())

        # If the user wants to use another device type, let them edit the value manually
        self.gpu_device_box.setEditable(True)

        layout.addWidget(self.enable_gpu_check, qt_pos_index, 0)
        layout.addWidget(self.gpu_device_box, qt_pos_index, 1)
        self.enable_gpu_check.stateChanged.connect(self.activate_enable_gpu_changed)

        qt_pos_index += 1
        self.frame_override_check = QCheckBox("Override Frame Range", self)
        self.frame_override_txt = QLineEdit(self)
        layout.addWidget(self.frame_override_check, qt_pos_index, 0)
        layout.addWidget(self.frame_override_txt, qt_pos_index, 1)
        self.frame_override_check.stateChanged.connect(self.activate_frame_override_changed)

        # Frame range validation
        # E.g.: 1-4,6,8,9-12
        # Note: ?: in regex groups all together as one result
        regex = QRegularExpression(
            r"\d+"  # unlimited numbers
            r"(?:-\d+)?"  # optional dash (-) and one or more digits
            r"(?:,(\s)?\d+"  # new parts split by commas (,) , allow 1 space for readability
            r"(?:-\d+)?)*"
        )  # can be repeated endlessly
        validator = QRegularExpressionValidator(regex, self.frame_override_txt)
        self.frame_override_txt.setValidator(validator)

        if self.dev_options:
            qt_pos_index += 1
            self.include_adaptor_wheels = QCheckBox(
                "Developer Option: Include Adaptor Wheels", self
            )
            layout.addWidget(self.include_adaptor_wheels, qt_pos_index, 0)

        layout.addItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding), 10, 0)

        self._fill_scenes_box()
        self._fill_cameras_box()
        self._fill_view_layers_box()

    def _fill_cameras_box(self):
        """Populate the cameras box with all available cameras."""
        saved_scene_name = self.scenes_box.currentData()
        selectable_cameras = blender_utils.get_renderable_cameras(saved_scene_name)
        _logger.info(f"filling cameras: {selectable_cameras}")
        with block_signals(self.cameras_box):
            # Save the current camera and reset the camera box list.
            saved_camera_name = self.cameras_box.currentData()
            self.cameras_box.clear()

            # Collect data and fill the combo box.
            self.cameras_box.addItem(
                COMBO_DEFAULT_ALL_RENDERABLE_CAMERAS, COMBO_DEFAULT_ALL_RENDERABLE_CAMERAS
            )
            for cam in selectable_cameras:
                self.cameras_box.addItem(cam, cam)

            # Re-select the camera if possible.
            index = self.cameras_box.findData(saved_camera_name)
            if index >= 0:
                self.cameras_box.setCurrentIndex(index)

    def _fill_scenes_box(self):
        """Populate the scenes box with all available scenes."""
        scenes = blender_utils.get_all_scenes()
        _logger.info(f"Filling scenes: {scenes}")
        with block_signals(self.scenes_box):
            # Save the current scene and reset the scene box list
            saved_scene_name = self.scenes_box.currentData()
            self.scenes_box.clear()
            for scene_name in scenes:
                self.scenes_box.addItem(scene_name, scene_name)

            # Re-select the Scene if possible.
            index = self.scenes_box.findData(saved_scene_name)
            if index >= 0:
                self.scenes_box.setCurrentIndex(index)

    def _fill_view_layers_box(self):
        """Populate the view layers box with all available view layers."""
        scene = self.scenes_box.currentData()
        _logger.info(f"Filling view layers for scene: {scene}")
        with block_signals(self.layers_box):
            # Save the current layer and reset the box list.
            saved_view_layer_name = self.layers_box.currentData()
            self.layers_box.clear()

            # Collect layer data to display.
            self.layers_box.addItem(
                COMBO_DEFAULT_ALL_RENDERABLE_LAYERS, COMBO_DEFAULT_ALL_RENDERABLE_LAYERS
            )
            for layer in blender_utils.get_renderable_view_layers(scene):
                self.layers_box.addItem(layer, layer)

            # Re-select the layer if possible
            index = self.layers_box.findData(saved_view_layer_name)
            if index >= 0:
                self.layers_box.setCurrentIndex(index)

    def _load(self, settings: BlenderSubmitterUISettings):
        """Set the UI to load sticky settings values."""
        self.proj_path_txt.setText(settings.project_path)
        self.op_dir_txt.set_text(settings.output_path)
        self.op_file_txt.setText(settings.output_file_prefix)

        i = self.scenes_box.findData(settings.scene_name)
        if i >= 0:
            self.scenes_box.setCurrentIndex(i)

        i = self.render_engine_box.findData(settings.renderer_name)
        if i >= 0:
            self.render_engine_box.setCurrentIndex(i)

        i = self.layers_box.findData(settings.view_layer_selection)
        if i >= 0:
            self.layers_box.setCurrentIndex(i)

        i = self.cameras_box.findData(settings.camera_selection)
        if i >= 0:
            self.cameras_box.setCurrentIndex(i)

        self.enable_gpu_check.setChecked(settings.enable_gpu)
        i = self.gpu_device_box.findData(settings.gpu_device)
        if i >= 0:
            self.gpu_device_box.setCurrentIndex(i)
        self.gpu_device_box.setEnabled(settings.enable_gpu)

        self.frame_override_check.setChecked(settings.override_frame_range)
        self.frame_override_txt.setEnabled(settings.override_frame_range)
        self.frame_override_txt.setText(settings.frame_list)

        if self.dev_options:
            self.include_adaptor_wheels.setChecked(settings.include_adaptor_wheels)

    def update_settings(self, settings: BlenderSubmitterUISettings):
        """Update a scene settings object with the latest values."""
        settings.project_path = self.proj_path_txt.text()
        settings.output_path = self.op_dir_txt.get_text()
        settings.output_file_prefix = self.op_file_txt.text()
        settings.scene_name = self.scenes_box.currentData()
        settings.renderer_name = self.render_engine_box.currentData()
        settings.ocio_config_path = ocio.get_ocio_path()
        settings.view_layer_selection = self.layers_box.currentData()
        settings.camera_selection = self.cameras_box.currentData()
        settings.enable_gpu = self.enable_gpu_check.isChecked()

        if settings.enable_gpu:
            # Blender's natively-supported options should be converted to uppercase from the menu options.
            # Custom options won't have associated data, so we use `currentText` to set the GPU device setting.
            settings.gpu_device = (
                self.gpu_device_box.currentData()
                if self.gpu_device_box.currentData()
                else self.gpu_device_box.currentText()
            )
        else:
            settings.gpu_device = "NONE"

        settings.override_frame_range = self.frame_override_check.isChecked()
        settings.frame_list = self.frame_override_txt.text()
        settings.include_adaptor_wheels = (
            self.include_adaptor_wheels.isChecked() if self.dev_options else False
        )

    def activate_enable_gpu_changed(self, state):
        """Set the activated/deactivated status of the GPU Device text box."""
        self.gpu_device_box.setEnabled(Qt.CheckState(state) == Qt.Checked)

    def activate_frame_override_changed(self, state):
        """Set the activated/deactivated status of the Frame override text box."""
        self.frame_override_txt.setEnabled(Qt.CheckState(state) == Qt.Checked)


class FileSearchLineEdit(QWidget):
    """Widget used to contain a line edit and a button which opens a file search box."""

    def __init__(self, directory_only=False, parent=None):
        super().__init__(parent=parent)

        self.directory_only = directory_only

        self.text_line = QLineEdit(self)

        self.button = QPushButton("...", parent=self)
        self.button.setMaximumSize(QSize(100, 40))
        self.button.clicked.connect(self._on_button_clicked)

        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.text_line)
        layout.addWidget(self.button)

    def _on_button_clicked(self):
        """Open a file picker to allow users to choose a file or directory."""
        if self.directory_only:
            new_txt = QFileDialog.getExistingDirectory(
                self,
                "Open Directory",
                self.get_text(),
                QFileDialog.ShowDirsOnly | QFileDialog.DontResolveSymlinks,
            )
        else:
            new_txt = QFileDialog.getOpenFileName(self, "Select File", self.get_text())

        if new_txt:
            self.set_text(new_txt)

    def set_text(self, txt: str) -> None:
        """Sets the text of the internal line edit."""
        self.text_line.setText(txt)

    def get_text(self) -> str:
        """Retrieves the text from the internal line edit."""
        return self.text_line.text()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import logging
from logging.handlers import RotatingFileHandler
from pathlib import Path
from typing import Optional

try:
    from deadline.client import config
except ImportError:
    pass

_DEFAULT_FORMATTER = logging.Formatter(
    fmt="%(asctime)s %(levelname)8s {%(threadName)-10s}:  %(module)s %(funcName)s: %(message)s",
)


def add_file_handler(
    file: Optional[Path] = None,
    logger: logging.Logger = logging.getLogger(),
    fmt: logging.Formatter = _DEFAULT_FORMATTER,
    level=logging.DEBUG,
):
    """Configure and add a file handler to the specified logger.

    If a handler with the same output file already exists on the logger, replace it.

    Args:
        file: The file to log to. Defaults to a file at `~/.deadline/logs/blender`.
        logger: The logger to add the handler to. Defaults to the root logger.
        fmt: The format string to use for the log messages on the new handler.
        level: The level to set the handler to. Defaults to DEBUG.
    """

    if file is None:
        file = Path.home() / ".deadline" / "logs" / "submitters" / "blender.log"
        file.parent.mkdir(parents=True, exist_ok=True)

    handler = RotatingFileHandler(file, maxBytes=10485760, backupCount=5)
    handler.setFormatter(fmt)
    handler.setLevel(level)

    # Find the handlers with the same file as the one we're adding, if any, and remove it.
    existing = []
    for h in logger.handlers:
        if isinstance(h, logging.FileHandler) and h.baseFilename == handler.baseFilename:
            existing.append(h)
    for h in existing:
        logger.removeHandler(h)
        logger.info(f"Removed existing file handler: {h}")

    logger.addHandler(handler)
    logger.info(f"Added file handler to handlers: {logger.handlers}")


def get_deadline_config_level():
    """Get the current log level set in the Deadline configuration file."""
    return config.config_file.get_setting("settings.log_level")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from __future__ import annotations

import logging
import os
import re
import sys
import threading
import time
from typing import Callable

from openjd.adaptor_runtime._version import version as openjd_adaptor_version
from openjd.adaptor_runtime.adaptors import Adaptor, AdaptorDataValidators, SemanticVersion
from openjd.adaptor_runtime.adaptors.configuration import AdaptorConfiguration
from openjd.adaptor_runtime.app_handlers import RegexCallback, RegexHandler
from openjd.adaptor_runtime.application_ipc import ActionsQueue, AdaptorServer
from openjd.adaptor_runtime.process import LoggingSubprocess
from openjd.adaptor_runtime_client import Action

from deadline.client.api import TelemetryClient, get_deadline_cloud_library_telemetry_client

from .._version import version as adaptor_version

_logger = logging.getLogger(__name__)


class BlenderNotRunningError(Exception):
    """Error that is raised when attempting
    to use Blender while it is not running"""


# Actions which must be queued before any others.
_FIRST_BLENDER_ACTIONS = ["scene_file", "gpu_device"]

# Order of execution is important.
_BLENDER_INIT_KEYS = [
    "render_scene",
    "view_layer",
    "camera",
    "output_dir",
    "output_file_name",
    "output_format",
]


def _check_for_exception(func: Callable) -> Callable:
    """
    Decorator that checks if an exception has been caught before calling the
    decorated function
    """

    def wrapped_func(self, *args, **kwargs):
        if not self._has_exception:  # Raises if there is an exception
            return func(self, *args, **kwargs)

    return wrapped_func


class BlenderAdaptor(Adaptor[AdaptorConfiguration]):
    """
    Adaptor that creates a session in Blender to Render interactively.
    """

    _SERVER_START_TIMEOUT_SECONDS = 30
    _SERVER_END_TIMEOUT_SECONDS = 30
    _BLENDER_START_TIMEOUT_SECONDS = 3600
    _BLENDER_END_TIMEOUT_SECONDS = 30

    _server: AdaptorServer | None = None
    _server_thread: threading.Thread | None = None
    _blender_client: LoggingSubprocess | None = None
    _action_queue = ActionsQueue()
    _is_rendering: bool = False
    # If a thread raises an exception we will update this to raise in the main thread
    _exc_info: Exception | None = None
    _performing_cleanup = False
    _telemetry_client: TelemetryClient | None = None
    _blender_version: str = ""

    @property
    def integration_data_interface_version(self) -> SemanticVersion:
        return SemanticVersion(major=0, minor=1)

    @staticmethod
    def _get_timer(timeout: int | float) -> Callable[[], bool]:
        """Given a timeout length, returns a lambda which returns True until the timeout occurs"""
        timeout_time = time.time() + timeout
        return lambda: time.time() < timeout_time

    @property
    def _has_exception(self) -> bool:
        """Property which checks the private _exc_info property for an exception

        Raises:
            self._exc_info: An exception if there is one

        Returns:
            bool: False there is no exception waiting to be raised
        """
        if self._exc_info and not self._performing_cleanup:
            raise self._exc_info
        return False

    @property
    def _blender_is_running(self) -> bool:
        """Property which indicates that the Blender client is running

        Returns:
            bool: True if the blender client is running, false otherwise
        """
        return self._blender_client is not None and self._blender_client.is_running

    @property
    def _blender_is_rendering(self) -> bool:
        """Property which indicates if blender is rendering

        Returns:
            bool: True if Blender is rendering, false otherwise
        """
        return self._blender_is_running and self._is_rendering

    @_blender_is_rendering.setter
    def _blender_is_rendering(self, value: bool) -> None:
        """Property setter which updates the private _is_rendering boolean.

        Args:
            value (bool): A boolean indicated if blender is rendering.
        """
        self._is_rendering = value

    def _wait_for_server(self) -> str:
        """
        Performs a busy wait for the server path that the adaptor server is running on, then
        returns it.

        Raises:
            RuntimeError: If the server does not finish initializing

        Returns:
            str: The server path the adaptor server is running on.
        """
        is_not_timed_out = self._get_timer(self._SERVER_START_TIMEOUT_SECONDS)
        while (self._server is None or self._server.server_path is None) and is_not_timed_out():
            time.sleep(0.01)

        if self._server is not None and self._server.server_path is not None:
            return self._server.server_path

        raise RuntimeError(
            "Could not find a server path because the server did not finish initializing"
        )

    def _start_blender_server(self) -> None:
        """
        Starts a server with the given ActionsQueue, attaches the server to
        the adaptor and serves forever in a blocking call.
        """
        self._server = AdaptorServer(self._action_queue, self)
        self._server.serve_forever()

    def _start_blender_server_thread(self) -> None:
        """
        Starts the Blender adaptor server in a thread.
        Sets the environment variable "BLENDER_ADAPTOR_SERVER_PATH" to the path the server is running
        on after the server has finished starting.
        """
        self._server_thread = threading.Thread(
            target=self._start_blender_server, name="BlenderAdaptorServerThread"
        )
        self._server_thread.start()
        os.environ["BLENDER_ADAPTOR_SERVER_PATH"] = self._wait_for_server()

    def _get_regex_callbacks(self) -> list[RegexCallback]:
        """
        Returns a list of RegexCallbacks used by the blender Adaptor

        Returns:
            list[RegexCallback]: List of Regex Callbacks to add
        """

        callback_list = []
        # Completion message is emitted by default_blender_handler.py::start_render
        completed_regexes = [re.compile("BlenderClient: Finished Rendering Frame [0-9]+")]
        progress_regexes = [
            re.compile(r"^Fra:.*Sample\s(\d+)\/(\d+)$"),  # Cycles
            re.compile(r"^Fra:.*Rendering\s(\d+)\s/\s(\d+)\ssamples$"),  # Eevee
            # Workbench renderer has no progress output
        ]
        error_regexes = [re.compile(".*Exception:.*|.*Error:.*|.*Warning.*")]
        # Capture the major minor patch version.
        version_regexes = [re.compile("BlenderClient: Blender Version ([0-9]+.[0-9]+.[0-9]+)")]

        callback_list.append(RegexCallback(completed_regexes, self._handle_complete))
        callback_list.append(RegexCallback(progress_regexes, self._handle_progress))
        if self.init_data.get("strict_error_checking", False):
            callback_list.append(RegexCallback(error_regexes, self._handle_error))
        callback_list.append(RegexCallback(version_regexes, self._handle_version))

        return callback_list

    @_check_for_exception
    def _handle_complete(self, match: re.Match) -> None:
        """
        Callback for stdout that indicate completeness of a render. Updates progress to 100
        Args:
            match (re.Match): The match object from the regex pattern that was matched the message
        """
        self._blender_is_rendering = False
        self.update_status(progress=100)

    @_check_for_exception
    def _handle_progress(self, match: re.Match) -> None:
        """
        Callback for stdout that indicate progress of a render.
        Args:
            match (re.Match): The match object from the regex pattern that was matched the message
        """
        capture_group = match.groups()
        progress = int(capture_group[0]) / int(capture_group[1]) * 100  # float
        progress = int(progress)  # Equivalent to math.floor
        self.update_status(progress=progress)

    def _handle_error(self, match: re.Match) -> None:
        """
        Callback for stdout that indicates an error or warning.
        Args:
            match (re.Match): The match object from the regex pattern that was matched the message

        Raises:
            RuntimeError: Always raises a runtime error to halt the adaptor.
        """
        self._exc_info = RuntimeError(f"Blender Encountered an Error: {match.group(0)}")

    def _handle_version(self, match: re.Match) -> None:
        """
        Callback for stdout that records the Blender version.
        Args:
            match (re.Match): The match object from the regex pattern that was matched the message
        """
        self._blender_version = match.groups()[0]

    @property
    def blender_client_path(self) -> str:
        """
        Obtains the blender_client.py path by searching directories in sys.path

        Raises:
            FileNotFoundError: If the blender_client.py file could not be found.

        Returns:
            str: The path to the blender_client.py file.
        """
        for dir_ in sys.path:
            path = os.path.join(
                dir_, "deadline", "blender_adaptor", "BlenderClient", "blender_client.py"
            )
            if os.path.isfile(path):
                return path
        raise FileNotFoundError(
            "Could not find blender_client.py. Check that the BlenderClient package is in one of the "
            f"following directories: {sys.path[1:]}"
        )

    def _start_blender_client(self) -> None:
        """
        Starts the blender client by launching Blender with the blender_client.py file.

        Blender and blenderPy must be on the system PATH, for example due to a Rez environment being active.

        Raises:
            FileNotFoundError: If the blender_client.py file could not be found.
        """
        blender_exe = os.environ.get("BLENDER_EXECUTABLE", "blender")
        regexhandler = RegexHandler(self._get_regex_callbacks())

        # Add the openjd namespace directory to PYTHONPATH, so that adaptor_runtime_client
        # will be available directly to the adaptor client.
        import openjd.adaptor_runtime_client

        from deadline import blender_adaptor

        openjd_namespace_dir = os.path.dirname(
            os.path.dirname(openjd.adaptor_runtime_client.__file__)
        )
        deadline_namespace_dir = os.path.dirname(os.path.dirname(blender_adaptor.__file__))
        python_path_addition = f"{openjd_namespace_dir}{os.pathsep}{deadline_namespace_dir}"
        if "PYTHONPATH" in os.environ:
            os.environ["PYTHONPATH"] = (
                f"{os.environ['PYTHONPATH']}{os.pathsep}{python_path_addition}"
            )
        else:
            os.environ["PYTHONPATH"] = python_path_addition

        if self.init_data["render_engine"] == "cycles":
            _logger.warning("Missing configuration for cycles render engine.")

        self._blender_client = LoggingSubprocess(
            args=[
                blender_exe,
                "--background",
                "--python",
                self.blender_client_path,
                "--python-use-system-env",
            ],
            stdout_handler=regexhandler,
            stderr_handler=regexhandler,
        )

    def _populate_action_queue(self) -> None:
        """
        Populates the adaptor server's action queue with actions from the init_data that the blender
        Client will request and perform. The action must be present in the _FIRST_blender_ACTIONS or
        _blender_INIT_KEYS set to be added to the action queue.
        """

        # Set up the renderer
        self._action_queue.enqueue_action(
            Action("render_engine", {"render_engine": self.init_data["render_engine"]})
        )

        for action_name in _FIRST_BLENDER_ACTIONS:
            self._action_queue.enqueue_action(self._action_from_action_item(action_name))

        for action_name in _BLENDER_INIT_KEYS:
            if action_name in self.init_data:
                _logger.info(f"ENQUEUING ACTION {action_name}")
                self._action_queue.enqueue_action(self._action_from_action_item(action_name))
            else:
                _logger.info(f"SKIPPING ACTION {action_name}: NOT IN INIT DATA")

    def on_start(self) -> None:
        """
        For job stickiness. Will start everything required for the Task.

        Raises:
            jsonschema.ValidationError: When init_data fails validation against the adaptor schema.
            jsonschema.SchemaError: When the adaptor schema itself is nonvalid.
            RuntimeError: If Blender did not complete initialization actions due to an exception
            TimeoutError: If Blender did not complete initialization actions due to timing out.
            FileNotFoundError: If the blender_client.py file could not be found.
        """
        cur_dir = os.path.dirname(__file__)
        schema_dir = os.path.join(cur_dir, "schemas")
        validators = AdaptorDataValidators.for_adaptor(schema_dir)
        validators.init_data.validate(self.init_data)

        self.update_status(progress=0, status_message="Initializing blender")
        self._start_blender_server_thread()
        self._populate_action_queue()

        self._start_blender_client()

        is_not_timed_out = self._get_timer(self._BLENDER_START_TIMEOUT_SECONDS)
        while (
            self._blender_is_running
            and not self._has_exception
            and len(self._action_queue) > 0
            and is_not_timed_out()
        ):
            time.sleep(0.1)  # busy wait for blender to finish initialization

        self._get_deadline_telemetry_client().record_event(
            event_type="com.amazon.rum.deadline.adaptor.runtime.start", event_details={}
        )

        if len(self._action_queue) > 0:
            if is_not_timed_out():
                raise RuntimeError(
                    "Blender encountered an error and was not able to complete initialization actions."
                )
            else:
                raise TimeoutError(
                    "Blender did not complete initialization actions in "
                    f"{self._BLENDER_START_TIMEOUT_SECONDS} seconds and failed to start."
                )

    def on_run(self, run_data: dict) -> None:
        """
        This starts a render in Blender for the given frame and performs a busy wait until the render
        completes.
        """
        if not self._blender_is_running:
            raise BlenderNotRunningError("Cannot render because Blender is not running.")

        # Run the camera action.
        self._action_queue.enqueue_action(Action("camera", run_data))

        # Load the validation schemas and validate the run data.
        cur_dir = os.path.dirname(__file__)
        schema_dir = os.path.join(cur_dir, "schemas")
        validators = AdaptorDataValidators.for_adaptor(schema_dir)
        validators.run_data.validate(run_data)

        # Queue the rendering tasks.
        self._blender_is_rendering = True
        self._action_queue.enqueue_action(Action("start_render", run_data))
        while self._blender_is_rendering and not self._has_exception:
            # Wait for the render to finish.
            time.sleep(0.1)

        if not self._blender_is_running and self._blender_client:
            # blender Client will always exist here.
            #  This is always an error case because the blender Client should still be running and
            #  waiting for the next command. If the thread finished, then we cannot continue
            exit_code = self._blender_client.returncode
            self._get_deadline_telemetry_client().record_error(
                {"exit_code": exit_code, "exception_scope": "on_run"}, str(RuntimeError)
            )
            raise RuntimeError(
                "Blender exited early and did not render successfully, please check render logs. "
                f"Exit code {exit_code}"
            )

    def on_stop(self) -> None:
        return

    def on_cleanup(self):
        """
        Cleans up the adaptor by closing the blender client and adaptor server.
        """
        self._performing_cleanup = True

        self._action_queue.enqueue_action(Action("close"), front=True)
        is_not_timed_out = self._get_timer(self._BLENDER_END_TIMEOUT_SECONDS)
        while self._blender_is_running and is_not_timed_out():
            time.sleep(0.1)
        if self._blender_is_running and self._blender_client:
            _logger.error(
                "Blender did not complete cleanup actions and failed to gracefully shutdown. "
                "Terminating."
            )
            self._blender_client.terminate()

        if self._server:
            self._server.shutdown()

        if self._server_thread and self._server_thread.is_alive():
            self._server_thread.join(timeout=self._SERVER_END_TIMEOUT_SECONDS)
            if self._server_thread.is_alive():
                _logger.error("Failed to shutdown the Blender Adaptor server.")

        self._performing_cleanup = False

    def on_cancel(self):
        """
        Cancels the current render if Blender is rendering.
        """
        _logger.info("CANCEL REQUESTED")
        if not self._blender_client or not self._blender_is_running:
            _logger.info("Nothing to cancel because Blender is not running")
            return

        # Terminate immediately since the Blender client does not have a graceful shutdown
        self._blender_client.terminate(grace_time_s=0)

    def _action_from_action_item(self, item_name: str) -> Action:
        """Shortcut to create an Action from the init_data."""
        return Action(
            item_name,
            {item_name: self.init_data[item_name]},
        )

    def _get_deadline_telemetry_client(self):
        """
        Wrapper around the Deadline Client Library telemetry client, in order to set package-specific information
        """
        if not self._telemetry_client:
            self._telemetry_client = get_deadline_cloud_library_telemetry_client()
            self._telemetry_client.update_common_details(
                {
                    "deadline-cloud-for-blender-adaptor-version": adaptor_version,
                    "blender-version": self._blender_version,
                    "open-jd-adaptor-runtime-version": openjd_adaptor_version,
                }
            )
        return self._telemetry_client



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""Functions to fill Blender job templates."""

from __future__ import annotations

import dataclasses
import json
import logging
import os
import traceback
from copy import deepcopy
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Optional

import yaml
from deadline.client.exceptions import DeadlineOperationError

_logger = logging.getLogger(__name__)

_ALL_CAMERAS = "ALL_CAMERAS"
_SETTINGS_FILE_EXT = ".deadline_render_settings.json"


@dataclass
class BlenderSubmitterUISettings:
    """Settings that the submitter UI will use."""

    name: str = field(default="Blender Submission", metadata={"sticky": True})
    description: str = field(default="", metadata={"sticky": True})

    submitter_name: str = field(default="Blender")
    renderer_name: str = field(default="cycles", metadata={"sticky": True})
    scene_name: str = field(default="Scene", metadata={"sticky": True})
    enable_gpu: bool = field(default=False, metadata={"sticky": True})
    gpu_device: str = field(default="None", metadata={"sticky": True})

    ui_group_label: str = field(default="Blender Settings")

    # Frame settings.
    # Whether to override the frame range.
    override_frame_range: bool = field(default=False, metadata={"sticky": True})
    # The frame range to use if override_frame_range_check is True.
    frame_list: str = field(default="1", metadata={"sticky": True})

    # Paths and files.
    project_path: str = field(default="")
    output_path: str = field(default="", metadata={"sticky": True})
    output_file_prefix: str = field(default="output_####", metadata={"sticky": True})
    input_filenames: list[str] = field(default_factory=list, metadata={"sticky": True})
    input_directories: list[str] = field(default_factory=list, metadata={"sticky": True})
    output_directories: list[str] = field(default_factory=list, metadata={"sticky": True})

    # OCIO config file is specified here.
    ocio_config_path: str = field(default="")

    # Layers settings.
    view_layer_selection: str = field(default="ViewLayer", metadata={"sticky": True})

    # Camera settings.
    all_layer_selectable_cameras: list[str] = field(default_factory=lambda: [_ALL_CAMERAS])
    current_layer_selectable_cameras: list[str] = field(default_factory=lambda: [_ALL_CAMERAS])
    camera_selection: str = field(default="Camera", metadata={"sticky": True})

    # Resolution settings.
    image_resolution: str = field(default="1920, 1080")

    # Parameter names.
    scene_name_parameter_name: str = field(default="RenderScene", metadata={"sticky": True})
    frames_parameter_name: str = field(default="Frames")
    output_file_prefix_parameter_name: str = field(default="OutputFileName")
    image_width_parameter_name: str = field(default="ResolutionX")
    image_height_parameter_name: str = field(default="ResolutionY")

    # developer options
    include_adaptor_wheels: bool = field(default=False, metadata={"sticky": True})

    def load_sticky_settings(self, scene: str) -> None:
        """Load and set sticky settings for the given scene.

        The path to the sticky settings file is determined by appending the `_SETTINGS_FILE_EXT` extension to the scene name.

        If the file does not exist, or if it is malformed, the default settings will be used instead.

        Args:
            scene: The name of the scene.
        """

        def warn(cause=None):
            traceback.print_exc()
            msg = f"Failed to load sticky settings from {sticky_file}."
            if cause:
                msg += " " + cause
            _logger.warn(msg)

        sticky_file = Path(scene).with_suffix(_SETTINGS_FILE_EXT)
        if not sticky_file.exists() or not sticky_file.is_file():
            warn("File does not exist.")
            return

        try:
            with open(sticky_file, encoding="utf8") as fh:
                sticky_settings = json.load(fh)
            if not isinstance(sticky_settings, dict):
                warn("Settings are not formatted as a dictionary.")
                return
            sticky_fields = {
                field.name: field
                for field in dataclasses.fields(self)
                if field.metadata.get("sticky")
            }
            for name, value in sticky_settings.items():
                if name in sticky_fields:
                    setattr(self, name, value)
        except (OSError, json.JSONDecodeError) as e:
            warn(str(e))

        _logger.info(f"Loaded sticky settings from {sticky_file}.")

    def save_sticky_settings(self, scene: str):
        """Save sticky settings to a file. Create or overwrite it if necessary, but don't create parent directories.

        The path to the sticky settings file is determined by appending the `_SETTINGS_FILE_EXT` extension to the scene name.
        """
        file = Path(scene).with_suffix(_SETTINGS_FILE_EXT)
        with open(file, "w", encoding="utf8") as fh:
            sticky_settings = {
                field.name: getattr(self, field.name)
                for field in dataclasses.fields(self)
                if field.metadata.get("sticky")
            }
            json.dump(sticky_settings, fh, indent=1)
        _logger.info(f"Saved sticky settings to {file}.")


@dataclass
class CommonLayerSettings:
    """Parameters common to a number of rendering layers."""

    renderer_name: str
    frame_range: str
    frames_parameter_name: Optional[str]
    renderable_camera_names: list[str]
    output_directories: str
    output_file_prefix: str
    output_file_prefix_parameter_name: Optional[str]
    ui_group_label: str
    image_width_parameter_name: Optional[str]
    image_height_parameter_name: Optional[str]
    image_resolution: tuple[int, int]
    scene_name: str


@dataclass
class Layer:
    """A unique rendering layer."""

    name: str
    common: CommonLayerSettings


def fill_job_template(
    settings: BlenderSubmitterUISettings,
    view_layers: list[Layer],
    host_requirements: Optional[dict],
):
    """Create and return a filled-in job template."""

    # Read the default job template.
    with open(Path(__file__).parent / "default_blender_template.yaml") as fh:
        job_template = yaml.safe_load(fh)

    job_template["name"] = settings.name
    if settings.description:
        job_template["description"] = settings.description

    # Use the first layer in the list as an example layer.
    # Aside from the `name` parameter, all parameters should be the same for all layers.
    layer_settings = view_layers[0].common

    # Add resolution parameters.
    job_template["parameterDefinitions"].append(
        {
            "name": layer_settings.image_width_parameter_name,
            "type": "INT",
            "userInterface": {
                "control": "SPIN_BOX",
                "label": "Image Width",
                "groupLabel": layer_settings.ui_group_label,
            },
            "minValue": 1,
            "description": "The image width.",
        }
    )
    job_template["parameterDefinitions"].append(
        {
            "name": layer_settings.image_height_parameter_name,
            "type": "INT",
            "userInterface": {
                "control": "SPIN_BOX",
                "label": "Image Height",
                "groupLabel": layer_settings.ui_group_label,
            },
            "minValue": 1,
            "description": "The image height.",
        }
    )

    # If an OCIO config file is set, we will have to add a job parameter and environment for it.
    if settings.ocio_config_path:
        _add_ocio_template_data(job_template)

    # For each layer, create a step based on the default step template.
    steps = []
    default_step_template = job_template["steps"][0]
    for layer in view_layers:
        step = _fill_step_template(layer, default_step_template, settings, host_requirements)
        _logger.info(f"Generated step definition for layer {layer.name}")
        steps.append(step)
    job_template["steps"] = steps

    # If this developer option is enabled, merge the adaptor_override_environment.
    if settings.include_adaptor_wheels:
        override_env_file = Path(__file__).parent / "adaptor_override_environment.yaml"
        with open(override_env_file) as f:
            override_environment = yaml.safe_load(f)

        err_msg = "The Developer Option 'Include Adaptor Wheels' is enabled, but "

        # Read DEVELOPMENT.md for instructions to create the wheels directory.
        wheels_path = Path(__file__).parent.parent.parent.parent / "wheels"
        if not wheels_path.exists() and wheels_path.is_dir():
            raise RuntimeError(
                err_msg + "the wheels directory does not exist:\n" + str(wheels_path)
            )

        wheels_path_package_names = {
            path.split("-", 1)[0] for path in os.listdir(wheels_path) if path.endswith(".whl")
        }
        expected = {"openjd_adaptor_runtime", "deadline", "deadline_cloud_for_blender"}
        if wheels_path_package_names != expected:
            raise RuntimeError(
                err_msg
                + "the wheels directory contains the wrong wheels:\n"
                + f"Expected: {expected}\n"
                + f"Actual: {wheels_path_package_names}"
            )

        override_adaptor_name_param = [
            param
            for param in override_environment["parameterDefinitions"]
            if param["name"] == "OverrideAdaptorName"
        ][0]
        override_adaptor_name_param["default"] = "blender-openjd"

        # There are no parameter conflicts between these two templates, so this works
        job_template["parameterDefinitions"].extend(override_environment["parameterDefinitions"])

        # Add the environment to the end of the template's job environments
        if "jobEnvironments" not in job_template:
            job_template["jobEnvironments"] = []
        job_template["jobEnvironments"].append(override_environment["environment"])

    return job_template


def _add_ocio_template_data(job_template: dict):
    job_template["parameterDefinitions"].append(
        {
            "name": "OCIOConfigPath",
            "type": "PATH",
            "objectType": "FILE",
            "dataFlow": "IN",
            "description": "The colour management info to render with.",
        }
    )

    if "jobEnvironments" not in job_template:
        job_template["jobEnvironments"] = []

    # Insert at index 0 so that OCIO is set before Blender runs
    job_template["jobEnvironments"].insert(
        0, {"name": "Set OCIO Path", "variables": {"OCIO": "{{Param.OCIOConfigPath}}"}}
    )


def _fill_step_template(
    layer: Layer,
    default_step_template: dict,
    settings: BlenderSubmitterUISettings,
    host_requirements: Optional[dict],
):
    """Return a defined step.

    Each step corresponds to a view layer in Blender. Its defined state is a dict that will be serialized to YAML and written to the job template. The default step template contains the basic structure of a step, but it is not filled in with the specific values for this layer.

    There are four steps to defining a step:
    - Update the step name to the name of the layer.
    - Update the init data.
    - Create a parameter space dimension for the selected frames.
    - Create a parameter space dimension for the selected cameras.
    - Inject the host requirements, if provided.

    NOTE: In f-strings, doubling { escapes it.
    """

    step = deepcopy(default_step_template)

    # Update the step name.
    step["name"] = layer.name

    def decode(bad_yaml: str) -> dict:
        """Some yaml inside the template cannot be parsed because it contains unescaped characters like {}. This function parses it manually.

        Parse a string like `'a: b\\nc: d\\ne: f'` to a dict `{'a': 'b', 'c': 'd', 'e': 'f'}`.
        """
        d = {}
        for line in bad_yaml.splitlines():
            key, value = line.split(": ", 1)
            d[key] = value
        return d

    def encode(d: dict) -> str:
        """Encode a dict `{'a': 'b', 'c': 'd', 'e': 'f'}` to a string `'a: b\nc: d\ne: f'`.

        The reverse operation of `decode`.
        """
        return "\n".join(f"{k}: {v}" for k, v in d.items())

    # Update the init data.
    # init_data["data"] is a string of key/values separated by newlines.
    # Parse it and update the values, then re-serialize it.
    updated_init_data = {
        "renderer": layer.common.renderer_name,
        "view_layer": layer.name,
        "output_file_prefix": "{{{{Param.{}}}}}".format(
            layer.common.output_file_prefix_parameter_name or "OutputFilePrefix"
        ),
        "image_width": "{{{{Param.{}}}}}".format(
            layer.common.image_width_parameter_name or "ImageWidth"
        ),
        "image_height": "{{{{Param.{}}}}}".format(
            layer.common.image_height_parameter_name or "ImageHeight"
        ),
    }
    init_data = step["stepEnvironments"][0]["script"]["embeddedFiles"][0]
    d = decode(init_data["data"])
    d.update(updated_init_data)
    init_data["data"] = encode(d)

    # Update the 'Param.Frames' reference in the Frame task parameter.
    param_defs = step["parameterSpace"]["taskParameterDefinitions"]
    if layer.common.frames_parameter_name:
        param_defs[0]["range"] = "{{{{Param.{}}}}}".format(layer.common.frames_parameter_name)

    # Create a parameter space dimension for the selected cameras.
    # Should be a list of strings, even if only one camera is selected.
    # NOTE this string var cross-references the constant defined in `scene_settings_widget.py`. Duplicated to avoid importing that file.
    if settings.camera_selection == "All Renderable Cameras":
        cameras = layer.common.renderable_camera_names
    else:
        cameras = [settings.camera_selection]
    param_defs.append({"name": "Camera", "type": "STRING", "range": cameras})
    run_data = step["script"]["embeddedFiles"][0]
    run_data["data"] += "camera: '{{Task.Param.Camera}}'\n"

    # If host requirements are provided, inject them into the step template.
    if host_requirements:
        step["hostRequirements"] = host_requirements
        _logger.debug(f"Injected host requirements into step {layer.name}: {host_requirements}")

    return step


def get_parameter_values(
    settings: BlenderSubmitterUISettings,
    layer_settings: CommonLayerSettings,
    queue_params: list[dict[str, Any]],
) -> list[dict[str, Any]]:
    """Collect the parameter values for the job bundle as a list of name/value dicts."""

    param_dict: dict[str, Any] = {
        "BlenderFile": settings.project_path,
        "OutputFileName": layer_settings.output_file_prefix,
        "OutputDir": layer_settings.output_directories,
        "RenderScene": layer_settings.scene_name,
        "RenderEngine": layer_settings.renderer_name,
        "GPUDevice": settings.gpu_device,
    }

    if layer_settings.frames_parameter_name:
        param_dict[layer_settings.frames_parameter_name] = layer_settings.frame_range
    if layer_settings.image_width_parameter_name:
        param_dict[layer_settings.image_width_parameter_name] = layer_settings.image_resolution[0]
    if layer_settings.image_height_parameter_name:
        param_dict[layer_settings.image_height_parameter_name] = layer_settings.image_resolution[1]

    # Format the parameter values as a list of dicts, each with a "name" and a "value" key.
    params = [{"name": k, "value": v} for k, v in param_dict.items()]

    # Check for any overlap between the job parameters we've defined and the queue parameters. This is an error, as we weren't synchronizing the values between the two different tabs where they came from.
    overlap: set[str] = {p["name"] for p in params}.intersection({p["name"] for p in queue_params})
    if overlap:
        raise DeadlineOperationError(
            "The following queue parameters conflict with the Blender job parameters:\n"
            + f"{', '.join(overlap)}"
        )

    # If specified, add the parameter value for OCIO here
    if settings.ocio_config_path:
        params.append({"name": "OCIOConfigPath", "value": settings.ocio_config_path})

    # If we're overriding the adaptor with wheels, remove the adaptor from the Packages
    if settings.include_adaptor_wheels:
        wheels_path = str(Path(__file__).parent.parent.parent.parent / "wheels")
        params.append({"name": "OverrideAdaptorWheels", "value": wheels_path})
        rez_param = {}
        conda_param = {}
        # Find the Packages parameter definition in the queue params.
        for param in queue_params:
            if param["name"] == "RezPackages":
                rez_param = param
            if param["name"] == "CondaPackages":
                conda_param = param

        # Remove the deadline_cloud_for_blender/blender-openjd rez package.
        if rez_param:
            rez_param["value"] = " ".join(
                pkg
                for pkg in rez_param["value"].split()
                if not pkg.startswith("deadline_cloud_for_blender")
            )
        if conda_param:
            conda_param["value"] = " ".join(
                pkg for pkg in conda_param["value"].split() if not pkg.startswith("blender-openjd")
            )

    params.extend({"name": param["name"], "value": param["value"]} for param in queue_params)

    return params



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from .default_blender_handler import DefaultBlenderHandler


class WorkbenchHandler(DefaultBlenderHandler):
    RENDER_ENGINE_NAME = "BLENDER_WORKBENCH"



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
"""Set of shortcut functions to query Blender scene data."""


from __future__ import annotations

import logging
import os
from pathlib import Path

import bpy

_logger = logging.getLogger(__name__)


def get_renderable_view_layers(saved_scene_name) -> list[str]:
    """Get the view layers associated with a scene that are selected to use
    during rendering.

    Args:
        saved_scene_name: The name of the scene.
    """
    scene_name = bpy.data.scenes[saved_scene_name].name
    layers = [layer.name for layer in bpy.data.scenes[scene_name].view_layers if layer.use]
    return layers


def get_renderable_cameras(saved_scene_name) -> list[str]:
    """Returns a list of all camera objects in the scene that are marked as renderable.

    Args:
        saved_scene_name: The name of the scene.
    """
    scene_name = bpy.data.scenes[saved_scene_name].name
    camera_names = [cam.name for cam in bpy.data.scenes[scene_name].objects if cam.type == "CAMERA"]
    return [cam for cam in camera_names if not bpy.data.objects[cam].hide_render]


def get_all_scenes() -> list[str]:
    """Returns a list of all scenes."""
    scene_names = [x.name for x in bpy.data.scenes]
    return scene_names


def get_scene_resolution(saved_scene_name):
    """Returns the resolution of the scene.

    Args:
        saved_scene_name: The name of the scene.
    """
    res_x = bpy.data.scenes[saved_scene_name].render.resolution_x
    res_y = bpy.data.scenes[saved_scene_name].render.resolution_y
    return res_x, res_y


def get_scene_name() -> str:
    """Construct and return a name for the current scene based on the currently opened `.blend` file."""
    scene_name = bpy.path.basename(bpy.context.blend_data.filepath).replace(".blend", "")
    return scene_name


def get_frames() -> str:
    """Returns the frame range of the active scene, formatted as a string e.g. `"1-10"`."""
    start = bpy.context.scene.frame_start
    end = bpy.context.scene.frame_end
    return str(start) + "-" + str(end)


def find_files(project_path, skip_temp=True, skip_nonexistent=True) -> list[Path]:
    """Returns a normalized list of paths to external files referenced by the loaded `.blend` file, augmented with `project_path`.

    Args:
        project_path: The path to the project directory.
        skip_temp: if True, skip all files from any of Blender's potential temp directories.
        skip_nonexistent: if True, skip all files that do not exist. When files are shared across machines, Blender may retain memory of original paths; this ensures that all retrieved paths exist on the local filesystem.
    """
    files = bpy.utils.blend_paths(absolute=True)
    files.append(project_path)
    files = set(Path(f) for f in files)

    if skip_temp:
        temp_dirs = _get_blender_temp_dirs()
        _logger.debug(f"Resolved Blender temp directories: {temp_dirs}")

        def is_in_temp(f: Path):
            """Returns True if the given file is in any of Blender's temp directories."""
            return any(f.is_relative_to(temp_dir) for temp_dir in temp_dirs)

        files = (f for f in files if not is_in_temp(f))

    if skip_nonexistent:
        files = (f for f in files if f.exists())

    return [f.resolve() for f in files]


def _get_blender_temp_dirs() -> list[Path]:
    """Returns a list of directories that Blender can try to use to store temporary files.

    Note that Blender only uses one of these at a time.

    Recreate the logic Blender uses to compute temp folders, as described here:
    https://docs.blender.org/manual/en/latest/advanced/blender_directory_layout.html#temporary-directory
    """
    dirs = []

    # `bpy.app.tempdir` seems to resolve to a project-specific directory, e.g. `'C:\\Users\\user\\AppData\\Local\\Temp\\blender_a07504\\'`. We want the parent directory, e.g. `'Temp\\'`.
    dirs.append(Path(bpy.app.tempdir).parent.resolve())

    # The user's preferences may specify a temp directory.
    user_pref_dir = bpy.context.preferences.filepaths.temporary_directory
    if user_pref_dir:
        dirs.append(Path(user_pref_dir).resolve())

    # System environment variables may specify a temp directory. Which one exists (if any) depends on the OS.
    for var in ["TEMP", "TMP", "TMP_DIR"]:
        if os.environ.get(var):
            dirs.append(Path(os.environ[var]).resolve())

    # The root temp directory is always a candidate.
    dirs.append(Path("/tmp").resolve())

    return list(set(dirs))



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from __future__ import annotations

import os
from types import FrameType
from typing import Optional

import bpy

# The blender Adaptor adds the `openjd` namespace directory to PYTHONPATH,
# so that importing just the adaptor_runtime_client should work.
try:
    from adaptor_runtime_client import ClientInterface
    from blender_adaptor.BlenderClient.render_handlers import get_render_handler
except (ImportError, ModuleNotFoundError):
    from openjd.adaptor_runtime_client import ClientInterface

    from deadline.blender_adaptor.BlenderClient.render_handlers import get_render_handler


class BlenderClient(ClientInterface):
    def __init__(self, server_path: str) -> None:
        super().__init__(server_path=server_path)
        print(f"BlenderClient: Blender Version {bpy.app.version_string}")
        self.actions.update({"render_engine": self.set_renderer})

    def set_renderer(self, renderer: dict):
        render_handler = get_render_handler(renderer["render_engine"])
        self.actions.update(render_handler.action_dict)

    def close(self, args: Optional[dict] = None) -> None:
        bpy.ops.wm.quit_blender()

    def graceful_shutdown(self, signum: int, frame: FrameType | None):
        bpy.ops.wm.quit_blender()


def main():
    server_path = os.environ.get("BLENDER_ADAPTOR_SERVER_PATH")
    if not server_path:
        raise OSError(
            "BlenderClient cannot connect to the Adaptor because the environment variable "
            "BLENDER_ADAPTOR_SERVER_PATH does not exist"
        )

    if not os.path.exists(server_path):
        raise OSError(
            "BlenderClient cannot connect to the Adaptor because the server at the path defined by "
            "the environment variable BLENDER_ADAPTOR_SERVER_PATH does not exist. Got: "
            f"{os.environ['BLENDER_ADAPTOR_SERVER_PATH']}"
        )

    client = BlenderClient(server_path)
    client.poll()


if __name__ == "__main__":  # pragma: no cover
    main()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from .blender_client import BlenderClient
from .render_handlers import DefaultBlenderHandler

__all__ = ["BlenderClient", "DefaultBlenderHandler"]



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from __future__ import annotations

import logging
import os
from typing import Optional

import bpy  # type: ignore

_logger = logging.getLogger(__name__)


class DefaultBlenderHandler:
    RENDER_ENGINE_NAME = "BLENDER_EEVEE_NEXT" if bpy.app.version >= (4, 2, 0) else "BLENDER_EEVEE"

    def __init__(self):
        """Initialize this handler."""

        # Define Blender actions to perform, each keyed to a callback function.
        # These cross-reference the Action instances created and queued in the Blender adaptor.
        self.action_dict = {
            "scene_file": self.set_scene_file,
            "gpu_device": self.set_gpu_device,
            "render_scene": self.set_render_scene,
            "view_layer": self.set_view_layer,
            "camera": self.set_camera,
            "output_dir": self.set_output_dir,
            "output_file_name": self.set_output_file_name,
            "output_format": self.set_output_format,
            "start_render": self.start_render,
        }

        # Set default values.
        self.camera_name = None
        self.scene_name = None
        self.output_dir = None
        self.output_file_name = None
        self.format = None
        self.view_layer_name = None

        _logger.debug(f"Initialized {self.__class__.__name__}")

    def _ensure_camera(self, data: dict) -> str:
        """
        Ensure that the camera provided in `data` exists in the scene and is renderable.
        Raises a RuntimeError otherwise.

        Returns the name of the camera.
        """
        camera: str = data.get("camera", self.camera_name)
        if camera is None:
            raise RuntimeError(f"No camera specified in data: {data}")

        # The ls function returns all cameras if they are set to renderable.
        scene_cameras = [
            cam.name for cam in bpy.data.scenes[self.scene_name].objects if cam.type == "CAMERA"
        ]
        if camera not in scene_cameras:
            raise RuntimeError(f"Camera {camera} does not exist.")

        if bpy.data.objects[camera].hide_render:
            raise RuntimeError(f"Camera {camera} is not renderable.")

        _logger.debug(f"Fetched camera to render: {camera}")
        return camera

    def _get_layer_to_render(self, data: dict) -> Optional[str]:
        """Gets the view_layer to render from the provided data.

        If no view_layer is specified in `data`, return None.
        """
        target = data.get("view_layer")
        if not target:
            _logger.debug("No view_layer specified in data.")
            return None

        # In Blender, set the right scene.
        scene = bpy.data.scenes[self.scene_name]
        bpy.context.window.scene = scene

        # Collect all layers that are set to render.
        enabled = {lyr.name for lyr in scene.view_layers if scene.view_layers[lyr.name].use}
        if target in enabled:
            _logger.debug(f"Fetched layer to render: {target}.")
            return target
        raise RuntimeError(
            f"view_layer {target} not found among available layers {sorted(enabled)}"
        )

    def start_render(self, data: dict) -> None:
        """
        Starts a render in Blender.

        Args:
            data (dict): The data given from the Adaptor. Keys expected: ['frame', 'camera']

        Raises:
            RuntimeError: If no camera was specified and no renderable camera was found
        """

        frame = data.get("frame")
        if frame is None:
            raise RuntimeError("BlenderClient: start_render called without a frame number.")

        bpy.context.scene.frame_start = frame
        bpy.context.scene.frame_end = frame

        # The camera fetched here is only for logging purposes.
        # The actual camera to render is set in the set_camera function,
        # and should have already been called as an Action.
        camera = self._ensure_camera(data)

        # Add layer and camera info to the output file path.
        bpy.context.scene.render.filepath = (
            f"{self.output_dir}/{self.view_layer_name}_{camera}_{self.output_file_name}"
        )
        _logger.debug(f"Set output file path to {bpy.context.scene.render.filepath}")

        _logger.debug(f"Rendering camera: {camera}")
        # The `animation` flag is required to correctly set the output file name.
        # See: https://docs.blender.org/api/current/bpy.ops.render.html#bpy.ops.render.render
        bpy.ops.render.render(animation=True, scene=self.scene_name)

        # This print statement (including flush) is required for Deadline to pick up successful task completion
        # See the Regex callbacks defined at `BlenderAdaptor/adaptor.py:_get_regex_callbacks`.
        print(f"BlenderClient: Finished Rendering Frame {frame}", flush=True)

    def set_camera(self, data: dict) -> None:
        """
        Sets the Camera that will be rendered.

        Args:
            data (dict): The data given from the Adaptor. Keys expected: ['camera']

        Raises:
            RuntimeError: If the camera is not renderable or does not exist
        """
        self.camera_name = data.get("camera")
        camera = self._ensure_camera(data)
        assert self.camera_name == camera, "Camera name mismatch: this should never happen."
        bpy.context.scene.camera = bpy.data.objects[camera]

    def set_render_scene(self, data: dict) -> None:
        """
        Sets the Scene that will be rendered and ensures it is set to the
        correct renderer.
        """
        self.scene_name = data.get("render_scene")
        bpy.context.window.scene = bpy.data.scenes[self.scene_name]
        _logger.debug(f"Set render engine: {self.RENDER_ENGINE_NAME}")
        bpy.context.scene.render.engine = self.RENDER_ENGINE_NAME

    def set_output_dir(self, data: dict) -> None:
        """
        Sets the output directory path.

        Args:
            data (dict): The data given from the Adaptor. Keys expected: ['output_dir']
        """
        output_dir = data.get("output_dir")
        if output_dir:
            self.output_dir = output_dir

    def set_output_file_name(self, data: dict) -> None:
        """
        Sets the output file name.

        Args:
            data (dict): The data given from the Adaptor. Keys expected: ['output_file_name']
        """
        output_file_name = data.get("output_file_name")
        _logger.debug(f"Set output file name: {output_file_name}")
        if output_file_name:
            self.output_file_name = output_file_name

    def set_output_format(self, data: dict) -> None:
        """
        Sets the output format.
        """
        format_ = data.get("output_format")
        _logger.debug(f"Set output format to {format_}")
        if format_:
            self.format = format_

    def set_view_layer(self, data: dict) -> None:
        """
        Sets the view_layer.

        For Blender, this means disable all layers except the one to render.

        Args:
            data (dict): The data given from the Adaptor. Keys expected: ['view_layer']

        Raises:
            RuntimeError: If the view_layer cannot be found
        """
        view_layer_name = self._get_layer_to_render(data)

        if view_layer_name is None:
            return

        for layer in bpy.context.window.scene.view_layers:
            layer.use = layer.name == view_layer_name
            _logger.debug(f"Set layer {layer.name} to render: {layer.use}")

        self.view_layer_name = view_layer_name

    def set_scene_file(self, data: dict):
        """Opens a Blender scene file.

        Args:
            data (dict): The data given from the Adaptor. Keys expected: ['scene_file']

        Raises:
            FileNotFoundError: If the file provided in the data dictionary does not exist.
        """
        file_path = data.get("scene_file", "")
        if not os.path.isfile(file_path):
            raise FileNotFoundError(f"The scene file '{file_path}' does not exist")
        bpy.ops.wm.open_mainfile(filepath=file_path)

    def set_gpu_device(self, data: dict) -> None:
        """
        Only implemented in the Cycles handler.
        """

        pass



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import os
from pathlib import Path

import PyOpenColorIO as ocio


def get_ocio_path() -> str:
    return os.environ["OCIO"] if "OCIO" in os.environ else ""


def get_ocio_config(filename: str) -> ocio.Config:
    return ocio.Config.CreateFromFile(filename)


def get_ocio_referenced_dirs(config: ocio.Config) -> list[str]:
    """
    Given an OCIO Config object, parse its `search_path` attribute and
    return a list of referenced directories.
    """

    path_list = []
    if config.getSearchPath():
        for path in config.getSearchPaths():
            # Paths may be absolute or relative to the config file.
            if not Path(path).is_absolute():
                path = Path.joinpath(Path(os.path.dirname(get_ocio_path())), path)

            path_list.append(path)

    return path_list



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from .default_blender_handler import DefaultBlenderHandler
from .cycles_blender_handler import CyclesHandler
from .workbench_blender_handler import WorkbenchHandler

__all__ = [
    "DefaultBlenderHandler",
    "CyclesHandler",
    "WorkbenchHandler",
    "get_render_handler",
]


def get_render_handler(renderer: str = "eevee") -> DefaultBlenderHandler:
    """
    Returns the render handler instance for the given renderer.

    Args:
        renderer (str, optional): The renderer to get the render handler of.
            Defaults to "Eevee".

    Returns the Render Handler instance for the given renderer.
    """
    if renderer == "eevee":
        return DefaultBlenderHandler()
    if renderer == "cycles":
        return CyclesHandler()
    if renderer == "workbench":
        return WorkbenchHandler()

    raise RuntimeError(f"Unsupported renderer: {renderer}")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import bpy
import logging

from .default_blender_handler import DefaultBlenderHandler

_logger = logging.getLogger(__name__)


class CyclesHandler(DefaultBlenderHandler):
    RENDER_ENGINE_NAME = "CYCLES"

    def set_gpu_device(self, data: dict) -> None:
        """
        Enables GPU rendering on the worker if specified in the user's configuration.

        Args:
            data (dict): The data given from the Adaptor. Keys expected: ['gpu_device']

        Raises:
            RuntimeError: If no devices of the specified type are available
        """

        gpu_device = data.get("gpu_device", "NONE")
        if gpu_device == "NONE":
            bpy.context.scene.cycles.device = "CPU"
            return

        # Supported GPU devices for Blender 3.6 and 4.2:
        # https://docs.blender.org/manual/en/3.6/render/cycles/gpu_rendering.html
        # https://docs.blender.org/manual/en/4.2/render/cycles/gpu_rendering.html
        GPU_DEVICE_TYPES_ITER = iter(["OPTIX", "CUDA", "HIP", "ONEAPI", "METAL"])

        cycles_preferences = bpy.context.preferences.addons["cycles"].preferences
        cycles_preferences.refresh_devices()

        # Attempt to set the user's chosen render device.
        # If it isn't available, loop through device types to find another GPU to use.
        devices = self._list_compatible_render_devices(cycles_preferences, gpu_device)
        while not devices and gpu_device != "NONE":
            _logger.warning(f'Device type "{gpu_device}" is not available on this worker.')
            gpu_device = next(GPU_DEVICE_TYPES_ITER, "NONE")
            _logger.warning(f'Attempting to use "{gpu_device}" instead.')
            devices = self._list_compatible_render_devices(cycles_preferences, gpu_device)

        bpy.context.scene.cycles.device = "GPU" if gpu_device != "NONE" else "CPU"
        cycles_preferences.compute_device_type = gpu_device
        _logger.debug(f"Set render device to {gpu_device}.")

    def _list_compatible_render_devices(self, preferences, gpu_device: str) -> list[str]:
        try:
            return preferences.get_devices_for_type(gpu_device)
        except ValueError as ve:
            _logger.warning(str(ve))
            return []



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Sanity checks done on submit or export bundle
"""
from typing import Union
from pathlib import Path

import bpy
from qtpy import QtWidgets

from . import blender_utils
from . import scene_settings_widget as ssw


def run_sanity_checks(settings, prompt_for_saving=True):
    # Prompt user for decision if there are unsaved changes
    if prompt_for_saving:
        _prompt_unsaved_changes()

    # Ensure there is a job name.
    if not settings.name:
        raise RuntimeError("No name was given for this submission")

    # Ensure that project path matches active file.
    if settings.project_path != bpy.context.blend_data.filepath:
        raise RuntimeError("Project Path does not match open file.")

    # Check if frame range is valid
    if settings.override_frame_range:
        _validate_frame_range(settings.frame_list)

    # Ensure the selected scene still exists
    if settings.scene_name not in blender_utils.get_all_scenes():
        raise RuntimeError(
            "The selected scene ({}) no longer exists! \n"
            "Please select a different scene.".format(settings.scene_name)
        )

    renderable_cameras = blender_utils.get_renderable_cameras(settings.scene_name)
    renderable_layers = blender_utils.get_renderable_view_layers(settings.scene_name)

    # Ensure there is at least one renderable camera.
    if len(renderable_cameras) == 0:
        raise RuntimeError(
            "There are no renderable cameras in the scene. Please enable at least one before submitting."
        )

    # Ensure the selected camera is still in the scene
    if (
        settings.camera_selection != ssw.COMBO_DEFAULT_ALL_RENDERABLE_CAMERAS
        and settings.camera_selection not in renderable_cameras
    ):
        raise RuntimeError(
            "The selected Camera ({}) is no longer in your scene! \n"
            "Please select a different Camera.".format(settings.camera_selection)
        )

    # Ensure the selected layer is still in the scene
    if (
        settings.view_layer_selection != ssw.COMBO_DEFAULT_ALL_RENDERABLE_LAYERS
        and settings.view_layer_selection not in renderable_layers
    ):
        raise RuntimeError(
            "The selected ViewLayer ({}) is no longer in your scene! \n"
            "Please select a different ViewLayer.".format(settings.view_layer_selection)
        )

    # Ensure the specified output directory exists.
    # If not, create it
    output_dir_as_path = Path(settings.output_path)
    if not (output_dir_as_path.exists()):
        try:
            output_dir_as_path.mkdir(parents=True)
        except Exception as e:
            raise RuntimeError(f"Could not create {output_dir_as_path}.\n\nError message:\n{e}")

    # Ensure the output file prefix was specified
    if not settings.output_file_prefix:
        raise RuntimeError("Please enter a prefix for the output files.")


def _prompt_unsaved_changes():
    """
    Prompts the user if the current file has unsaved changes.

    User can either:

    * Save (and continue) or;
    * Don't save (and continue) or;
    * Cancel

    Cancelling raises a RuntimeError.
    """
    # Check if file is saved before submit, if not add prompt to save
    if not bpy.data.is_dirty:
        return
    # Show message to user for decision
    msg = QtWidgets.QMessageBox()
    msg.setText("Scene has unsaved changes.")
    msg.setWindowTitle("Blender")
    msg.addButton("Save", QtWidgets.QMessageBox.YesRole)
    msg.addButton("Don't Save", QtWidgets.QMessageBox.NoRole)
    msg.addButton(QtWidgets.QMessageBox.Cancel)
    return_value = msg.exec()
    if return_value == 0:
        bpy.ops.wm.save_mainfile()
    elif return_value == QtWidgets.QMessageBox.Cancel:
        raise RuntimeError("Submission cancelled.")


def _validate_frame_range(frames: str):
    """
    Validates the given frame range string.

    Raises a RuntimeError if it is invalid.

    Example inputs:

        "1"
        "1,3,5-7"
        "1,3,3,3,5" will result in a RuntimeError.
    """
    if not frames:
        raise RuntimeError("Override Frame Range checked but no frame range was given")
    if not is_correct_frame_range(frames):
        raise RuntimeError(
            "You entered an invalid frame range. Please make sure that the first number in the range "
            "is smaller than the second number. \n"
            "E.g.: 10-5 is invalid, 5-10 is valid."
        )
    if not has_no_duplicate_frames(frames)[0]:
        raise RuntimeError(
            "You entered an invalid frame range. Please make sure there are no duplicate frames in "
            "your range. \n"
            f"Duplicate frames: {has_no_duplicate_frames(frames)[1]}"
        )


def is_correct_frame_range(frames: str) -> bool:
    """
    Check if first number in range is smaller than the second.
    Input from text field can only have numbers, commas or dashes due to
    regex validator on field.

    :param frames: frame range you want to check
    :type frames: str
    :returns: boolean that indicates whether the frame range was valid
    :return_type: bool
    """
    frames_input = frames.strip()
    override_frames = frames_input.split(",")
    # only need to check for ranges, not single frames
    frame_ranges = [pair for pair in override_frames if "-" in pair]
    for range_ in frame_ranges:
        numbers = range_.split("-")
        if int(numbers[1]) <= int(numbers[0]):
            return False
    return True


def has_no_duplicate_frames(frames: str) -> list[Union[bool, str]]:
    """
    Check if there are any repeating numbers in the given frame range.
    Input from text field can only have numbers, commas or dashes due to
    regex validator on field.

    :param frames: frame range you want to check
    :type frames: str
    :returns: a list containing a boolean that indicates whether there were no duplicates and a string with the
    duplicate numbers
    :return_type: list[Union[bool, str]]
    """
    # remove any spaces and split into groups
    frames_input = frames.strip()
    override_frames = frames_input.split(",")
    frames_to_render: list[int] = []
    has_no_duplicates = True
    duplicates = []
    for frames in override_frames:
        try:
            if frames_to_render:
                if int(frames) in frames_to_render:
                    has_no_duplicates = False
                    duplicates.append(frames)
            frames_to_render.append(int(frames))
        # when there is a dash in the frame string it can't be converted to an int
        # this way we can easily split single frames from ranges
        except ValueError:
            numbers = frames.split("-")
            for i in range(int(numbers[0]), int(numbers[1]) + 1):
                if frames_to_render:
                    if i in frames_to_render:
                        has_no_duplicates = False
                        duplicates.append(str(i))
                frames_to_render.append(i)
    return [has_no_duplicates, ", ".join(duplicates)]



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import os
from pathlib import Path
from typing import Any, Optional

import bpy
from qtpy.QtCore import Qt  # type: ignore

from deadline.client import api
from deadline.client.job_bundle._yaml import deadline_yaml_dump
from deadline.client.job_bundle.submission import AssetReferences
from deadline.client.ui.dialogs.submit_job_to_deadline_dialog import SubmitJobToDeadlineDialog

from . import blender_utils as bu
from . import sanity_checks as sc
from . import scene_settings_widget as ssw
from . import template_filling as tf
from . import ocio_utils as ocio
from ._version import version
from ._version import version_tuple as adaptor_version_tuple


def create_deadline_dialog(parent=None) -> SubmitJobToDeadlineDialog:
    """Main function that shows the UI.

    Initialize job settings with default values or values from the scene. Build and return a dialogue widget with these settings.

    This function is called every time the `Render > Submit to AWS Deadline Cloud` window is clicked; there is no persistence except for the sticky settings.
    """

    # The scene must be saved before opening the submit dialog.
    if not bpy.data.is_saved:
        raise RuntimeError(
            "The Blender scene is not saved to disk. Please save it before opening the submitter dialog."
        )

    # Initialize telemetry client, opt-out is respected
    api.get_deadline_cloud_library_telemetry_client().update_common_details(
        {
            "deadline-cloud-for-blender-submitter-version": version,
            "blender-version": bpy.app.version_string,
        }
    )

    # Initialize render settings with default values.
    settings = tf.BlenderSubmitterUISettings()

    # Set settings dependent on scene.
    settings.name = bu.get_scene_name()
    settings.project_path = bpy.context.blend_data.filepath
    settings.frame_list = bu.get_frames()

    # For the output path, first check for a value in the Scene settings
    if os.path.dirname(bpy.context.scene.render.filepath):
        settings.output_path = os.path.dirname(bpy.path.abspath(bpy.context.scene.render.filepath))
    # If none, use the one in Preferences
    elif bpy.context.preferences.filepaths.render_output_directory:
        settings.output_path = bpy.context.preferences.filepaths.render_output_directory
    # If neither of these are set, use the job bundle directory by default
    else:
        settings.output_path = os.path.dirname(bpy.context.blend_data.filepath)

    if bpy.path.basename(bpy.context.scene.render.filepath):
        settings.output_file_prefix = bpy.path.basename(bpy.context.scene.render.filepath)

    # Read the user's preferences to set GPU settings.
    settings.enable_gpu = bpy.context.scene.cycles.device == "GPU"

    if bpy.context.preferences.addons["cycles"].preferences.compute_device_type != "NONE":
        settings.gpu_device = bpy.context.preferences.addons[
            "cycles"
        ].preferences.compute_device_type

    # Load and set sticky settings, if any.
    settings.load_sticky_settings(settings.project_path)

    settings.current_layer_selectable_cameras = [settings.camera_selection]
    settings.all_layer_selectable_cameras = [settings.camera_selection]

    # Set auto-detected attachments.
    auto_detected_attachments = _get_auto_detected_assets(settings.project_path)

    # If the user has an OCIO config file set, we will have to upload any directories referenced by that file as job attachments.
    if ocio.get_ocio_path():
        ocio_config = ocio.get_ocio_config(ocio.get_ocio_path())
        for path in ocio.get_ocio_referenced_dirs(ocio_config):
            auto_detected_attachments.input_directories.add(str(path))

    # Set regular attachments.
    attachments = AssetReferences(
        input_filenames=set(settings.input_filenames),
        input_directories=set(settings.input_directories),
        output_directories=set(settings.output_directories),
    )

    # https://docs.blender.org/api/current/bpy.app.html#bpy.app.version
    blender_version = ".".join(map(str, bpy.app.version[:2]))
    adaptor_version = ".".join(str(v) for v in adaptor_version_tuple[:2])
    # Need Blender and the Blender OpenJD application interface adaptor
    rez_packages = f"blender-{blender_version} deadline_cloud_for_blender"
    conda_packages = f"blender={blender_version}.* blender-openjd={adaptor_version}.*"

    # Create and return the dialog widget.
    # dialog = SubmitJobToDeadlineDialog(
    dialog = SubmitJobToDeadlineDialog(
        job_setup_widget_type=ssw.SceneSettingsWidget,
        initial_job_settings=settings,
        initial_shared_parameter_values={
            "RezPackages": rez_packages,
            "CondaPackages": conda_packages,
        },
        auto_detected_attachments=auto_detected_attachments,
        attachments=attachments,
        on_create_job_bundle_callback=_create_bundle,
        parent=parent,
        f=Qt.Tool,
        show_host_requirements_tab=True,
    )
    dialog.setWindowFlags(Qt.WindowStaysOnTopHint)
    return dialog


def _create_bundle(
    widget: SubmitJobToDeadlineDialog,
    job_bundle_dir: str,
    settings: tf.BlenderSubmitterUISettings,
    queue_parameters: list[dict[str, Any]],
    asset_references: AssetReferences,
    requirements: Optional[dict[str, Any]] = None,
    purpose: Optional[str] = None,
) -> None:
    """Create and write Deadline job bundle files to the given directory. Also save sticky settings.

    NOTE: The parameters to this function are fixed by the Deadline API. Do not change them, even if they are unused. See the source of the `SubmitJobToDeadlineDialog` class.

    This function is called when the user clicks the "Submit" button in the submitter dialog.

    Args:
        widget: The submitter dialog widget.
        job_bundle_dir: The directory to write the job bundle files to.
        settings: The job settings.
        queue_parameters: The queue parameters.
        asset_references: The asset references.
        requirements: The host requirements. There are only passed if the user has specified custom host requirements in the UI. If "all available worker hosts" are selected, this is None.
        purpose: The purpose of the job bundle.
    """
    _create_bundle_internal(
        widget, job_bundle_dir, settings, queue_parameters, asset_references, requirements, purpose
    )


def _create_bundle_internal(
    widget: SubmitJobToDeadlineDialog,
    job_bundle_dir: str,
    settings: tf.BlenderSubmitterUISettings,
    queue_parameters: list[dict[str, Any]],
    asset_references: AssetReferences,
    requirements: Optional[dict[str, Any]] = None,
    purpose: Optional[str] = None,
    prompt_for_saving=True,
) -> None:
    """Create and write Deadline job bundle files to the given directory. Also save sticky settings.
    The internal variation exists so that the functionality can be accessed without a UI.

    Args:
        widget: The submitter dialog widget.
        job_bundle_dir: The directory to write the job bundle files to.
        settings: The job settings.
        queue_parameters: The queue parameters.
        asset_references: The asset references.
        requirements: The host requirements. There are only passed if the user has specified custom host requirements in the UI. If "all available worker hosts" are selected, this is None.
        purpose: The purpose of the job bundle.
    """
    # Make sure the output path used for submission is absolute
    settings.output_path = bpy.path.abspath(settings.output_path)

    # Run sanity checks on submission
    sc.run_sanity_checks(settings, prompt_for_saving)

    renderable_cameras = bu.get_renderable_cameras(settings.scene_name)

    common_layer_settings = tf.CommonLayerSettings(
        renderer_name=settings.renderer_name,
        frame_range=bu.get_frames(),
        renderable_camera_names=renderable_cameras,
        output_directories=settings.output_path,
        output_file_prefix=settings.output_file_prefix,
        image_resolution=(
            bpy.context.scene.render.resolution_x,
            bpy.context.scene.render.resolution_y,
        ),
        ui_group_label=settings.ui_group_label,
        frames_parameter_name=settings.frames_parameter_name,
        output_file_prefix_parameter_name=settings.output_file_prefix_parameter_name,
        image_width_parameter_name=settings.image_width_parameter_name,
        image_height_parameter_name=settings.image_height_parameter_name,
        scene_name=settings.scene_name,
    )

    # Add selected layers to the list of layers to render.
    layers: list[tf.Layer] = []
    if settings.view_layer_selection == ssw.COMBO_DEFAULT_ALL_RENDERABLE_LAYERS:
        for layer in bu.get_renderable_view_layers(settings.scene_name):
            layers.append(tf.Layer(layer, common_layer_settings))
    else:
        layers.append(tf.Layer(settings.view_layer_selection, common_layer_settings))

    if settings.override_frame_range:
        common_layer_settings.frame_range = settings.frame_list

    job_template = tf.fill_job_template(settings, layers, requirements)
    parameter_values = tf.get_parameter_values(settings, common_layer_settings, queue_parameters)

    # Write the job bundle files.
    job_bundle_path = Path(job_bundle_dir)
    for file_name, to_dump in [
        ("template.yaml", job_template),
        ("parameter_values.yaml", {"parameterValues": parameter_values}),
        ("asset_references.yaml", asset_references.to_dict()),
    ]:
        with open(job_bundle_path / file_name, "w", encoding="utf8") as f:
            deadline_yaml_dump(to_dump, f, indent=1)

    # save sticky settings
    attachments: AssetReferences = widget.job_attachments.attachments
    settings.input_filenames = sorted(attachments.input_filenames)
    settings.input_directories = sorted(attachments.input_directories)
    settings.input_filenames = sorted(attachments.input_filenames)
    scene_filename = settings.project_path
    settings.save_sticky_settings(scene_filename)


def _get_auto_detected_assets(project_path: str) -> AssetReferences:
    files = bu.find_files(project_path)

    # Sort auto-detected attachments to classify files and directories correctly.
    input_filenames = set()
    input_directories = set()
    for f in files:
        if f.is_dir():
            input_directories.add(str(f))
        else:
            input_filenames.add(str(f))

    return AssetReferences(input_filenames=input_filenames, input_directories=input_directories)


if __name__ == "__main__":
    create_deadline_dialog()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Registration of Deadline Cloud Submitter Addon + activate logger
"""
import logging
from pathlib import Path
import subprocess
import sys

import bpy  # noqa
from bpy.types import Operator

from . import logutil

# NOTE: Variables are NOT allowed to be in bl_info since
#       blender parses this __init__.py source for this variable
#       and is not aware of any additional context.
#       ref: https://developer.blender.org/docs/handbook/addons/addon_meta_info/
bl_info = {
    "name": "Deadline Cloud for Blender",
    "description": "Submit to AWS Deadline Cloud",
    "author": "AWS",
    "version": (0, 5, 2),
    "blender": (3, 6, 0),
    "category": "Render",
}

# Configure logging.
logutil.add_file_handler()

_logger = logging.getLogger(__name__)

addon_keymaps = []


class DEADLINE_CLOUD_OT_open_dialog(Operator):
    """Open deadline cloud dialog."""

    bl_idname = "ops.open_deadline_cloud_dialog"
    bl_label = "AWS Deadline Cloud"
    bl_options = {"REGISTER"}

    # Execution after the window was closed with ok button.
    def execute(self, context):
        """Execute the operator.

        See the bpy Operator docs: https://docs.blender.org/api/current/bpy.types.Operator.html
        """
        if not self.has_gui_deps():
            self.install_gui()

        from qtpy import QtCore, QtWidgets
        from deadline_cloud_blender_submitter.open_deadline_cloud_dialog import (
            create_deadline_dialog,
        )

        self.app = QtWidgets.QApplication.instance()
        if not self.app:
            self.app = QtWidgets.QApplication(sys.argv)

        try:
            # optionally use the blender_stylesheet if it exists
            import blender_stylesheet

            blender_stylesheet.setup()
        except ImportError:
            _logger.info("blender_stylesheet package is not available. Skipping")

        _logger.info("Initializing Deadline Cloud Blender Submitter UI")
        try:
            self.widget = create_deadline_dialog()
        except RuntimeError as e:
            self.report({"ERROR"}, str(e))
            return {"CANCELLED"}

        # Set window flags.
        flag_info = f"Current window flags: {bin(int(self.widget.windowFlags()))}"
        _logger.debug(flag_info)
        # Option 0: Default behaviour: window will always stay on top of other windows.
        # self.widget.setWindowFlags(self.widget.windowFlags() | QtCore.Qt.WindowStaysOnTopHint)
        # Option 1: go into background when focus is lost, keep a taskbar entry.
        self.widget.setWindowFlags(self.widget.windowFlags() & ~QtCore.Qt.WindowStaysOnTopHint | QtCore.Qt.WindowCloseButtonHint)  # type: ignore
        # Option 2: tool window.
        # Setting the window to be a tool window has the desired effect of making it stay on top
        # of the Blender window only. But the disadvantage that, without a presence in the menu bar,
        # it gets lost when focusing away from Blender.
        # self.widget.setWindowFlags(self.widget.windowFlags() & ~QtCore.Qt.WindowStaysOnTopHint)
        # self.widget.setWindowFlags(self.widget.windowFlags() | QtCore.Qt.Tool)
        _logger.debug(flag_info)

        self.widget.exec_()
        self.report({"INFO"}, "OK!")
        return {"FINISHED"}

    def draw(self, context):
        layout = self.layout
        col = layout.column()
        col.label(text="Press 'OK' to install GUI dependencies. Please wait...")

    def invoke(self, context, event):
        if self.has_gui_deps():
            # don't prompt user if gui deps exist
            return self.execute(context)
        wm = context.window_manager
        return wm.invoke_props_dialog(self)

    def has_gui_deps(self):
        try:
            import qtpy  # noqa
            from deadline_cloud_blender_submitter.open_deadline_cloud_dialog import (  # noqa
                create_deadline_dialog,
            )
        except Exception as e:
            # qtpy throws a QtBindingsNotFoundError when running
            # from qtpy import QtBindingsNotFoundError
            if not (type(e).__name__ == "QtBindingsNotFoundError" or isinstance(e, ImportError)):
                raise
            return False

        return True

    def install_gui(self):
        import deadline.client

        deadline.client.version
        pip_install_command = [
            sys.executable,
            "-m",
            "pip",
            "install",
            f"deadline[gui]=={deadline.client.version}",
        ]
        # module_directory assumes relative install location of:
        #   * [installdir]/Submitters/Blender/python/addons/deadline_cloud_blender_submitter/__init__.py
        #   * [installdir]/Submitters/Blender/python/modules/
        module_directory = Path(__file__).parent.parent.parent / "modules"
        if module_directory.exists():
            _logger.info(f"Missing GUI libraries, installing deadline[gui] to {module_directory}")
            pip_install_command.extend(["--target", str(module_directory)])
        else:
            _logger.info(
                "Missing GUI libraries with non-standard set-up, installing deadline[gui] into Blender's python"
            )

        subprocess.run(pip_install_command)


def deadline_cloud_dialog_topbar_btn(self, context):
    """Deadline Cloud Dialog button."""
    self.layout.separator()
    self.layout.operator(
        DEADLINE_CLOUD_OT_open_dialog.bl_idname, text="Submit to AWS Deadline Cloud"
    )


def register():
    """Register the addon to the Blender UI."""
    bpy.utils.register_class(DEADLINE_CLOUD_OT_open_dialog)
    bpy.types.TOPBAR_MT_render.append(deadline_cloud_dialog_topbar_btn)

    # keymapping for pop-out window
    wm = bpy.context.window_manager
    kc = wm.keyconfigs.addon
    if kc:
        km = kc.keymaps.new(name="3D View", space_type="VIEW_3D")
        kmi = km.keymap_items.new(
            "ops.open_deadline_cloud_dialog", type="F", value="PRESS", shift=True
        )
        addon_keymaps.append((km, kmi))


def unregister():
    """Unregister the addon from the Blender UI."""
    for km, kmi in addon_keymaps:
        km.keymap_items.remove(kmi)
    addon_keymaps.clear()
    bpy.utils.unregister_class(DEADLINE_CLOUD_OT_open_dialog)
    bpy.types.TOPBAR_MT_render.remove(deadline_cloud_dialog_topbar_btn)


if __name__ == "__main__":
    register()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
This file contains the AddonPreferences for the deadline submitter ui.
"""

from bpy.props import BoolProperty, EnumProperty, StringProperty
from bpy.types import AddonPreferences


# Addon preferences / layout
class DEADLINE_CLOUD_AddonPreferences(AddonPreferences):
    """NOTE: not currently used."""

    # this must match the add-on name, use '__package__'
    # when defining this in a submodule of a python package.
    bl_idname = __package__

    # TODO: Will need to be a dynamic selection

    job_history: StringProperty(  # type: ignore
        name="Job History Dir",
        default="",
    )
    user_identities: EnumProperty(  # type: ignore
        name="User Identities",
        description="User Identities",
        items=[("op1", "False", ""), ("op2", "True", "")],
    )
    auto_accept_confirm_prompt: BoolProperty(name="Auto Accept Confirmation Prompts")  # type: ignore
    current_logging_level: EnumProperty(  # type: ignore
        name="Current Logging Level",
        description="Current Logging Level",
        items=[("op1", "Debug", ""), ("op2", "None", "")],
    )

    def draw(self, context):
        """
        This draw builds the AddonPreferences ui in
        settings > preferences > addons > Deadline Cloud Submitter Addon
        """
        layout = self.layout
        layout.label(text="This is a preferences view for our add-on")
        box = layout.box()
        box.label(text="Global Settings:")
        row = box.row()
        split = row.split(factor=0.9)
        split.prop(context.scene, "aws_profile")
        split.operator("ops.dlcloud_add_to_aws_profile", icon="FILE_REFRESH")

        box = layout.box()
        box.label(text="Profile Settings:")
        row = box.row()
        split = row.split(factor=0.9)
        split.prop(self, "job_history")
        split.operator("ops.dlcloud_get_job_history_dir", text="", icon="FOLDER_REDIRECT")
        row = box.row()
        split = row.split(factor=0.9)
        split.prop(context.scene, "default_farm")
        split.operator("ops.dlcloud_add_to_default_farm", icon="FILE_REFRESH")
        row = box.row()
        row.prop(self, "user_identities")

        box = layout.box()
        box.label(text="Farm Settings:")
        row = box.row()
        split = row.split(factor=0.9)
        split.prop(context.scene, "default_queue")
        split.operator("ops.dlcloud_add_to_default_queue", icon="FILE_REFRESH")

        box = layout.box()
        box.label(text="General Settings:")
        row = box.row()
        row.prop(self, "auto_accept_confirm_prompt")
        row = box.row()
        row.prop(self, "current_logging_level")
